from typing import Optional, List, Union

from django_scoped_permissions.core import scopes_grant_permissions
from django_scoped_permissions.util import expand_scopes_from_context


class ScopedPermissionRequirement:
    """
    ScopedPermissionRequirement describes a single permission requirement. A ScopedPermissionGuard
    holds one or more such requirements to describe a complex permission guard.
    """

    def __init__(self, scope: str, verb: Optional[str] = None):
        self.scope = scope
        self.verb = verb


def _evaluate_value(value, granting_scopes: List[str], context=None):
    if not context:
        context = {}

    granting_scopes = expand_scopes_from_context(granting_scopes, context)

    if isinstance(value, SPRBinOp) or isinstance(value, SPRUnOp):
        return value.has_permission(granting_scopes, context)
    elif isinstance(value, str):
        required_scopes = expand_scopes_from_context([value], context)
        return scopes_grant_permissions(required_scopes, granting_scopes)
    elif isinstance(value, list):
        required_scopes = expand_scopes_from_context(value, context)
        return scopes_grant_permissions(required_scopes, granting_scopes)
    elif isinstance(value, ScopedPermissionRequirement):
        required_scopes = expand_scopes_from_context([value.scope], context)
        return scopes_grant_permissions(required_scopes, granting_scopes, value.verb)
    elif isinstance(value, bool):
        return value
    else:
        return False


class SPRUnOp:
    def __init__(self, value):
        self.value = value

    def has_permission(self, granting_scopes: List[str], context=None):
        return _evaluate_value(self.value, granting_scopes, context)


class SPRBinOp:
    def __init__(self, lhs, rhs):
        self.lhs = lhs
        self.rhs = rhs

    def has_permission(self, granting_scopes: List[str], context=None):
        return False

    def lhs_has_permission(self, granting_scopes: List[str], context=None):
        return _evaluate_value(self.lhs, granting_scopes, context)

    def rhs_has_permission(self, granting_scopes: List[str], context=None):
        return _evaluate_value(self.rhs, granting_scopes, context)


class SPRAnd(SPRBinOp):
    def has_permission(self, granting_scopes: List[str], context=None):
        return self.lhs_has_permission(granting_scopes) & self.rhs_has_permission(
            granting_scopes
        )


class SPROr(SPRBinOp):
    def has_permission(self, granting_scopes: List[str], context=None):
        return self.lhs_has_permission(
            granting_scopes, context
        ) | self.rhs_has_permission(granting_scopes, context)


class SPRXor(SPRBinOp):
    def has_permission(self, granting_scopes: List[str], context=None):
        return self.lhs_has_permission(
            granting_scopes, context
        ) ^ self.rhs_has_permission(granting_scopes, context)


class SPRNot(SPRUnOp):
    def __init__(self, value):
        self.value = value

    def has_permission(self, granting_scopes: List[str], context=None):
        return not super().has_permission(granting_scopes, context)


class ScopedPermissionGuard:
    """
    ScopedPermissionGuard represents a complex permission guard, guarding some resource.

    The guard may contain a single permission requirement, or multiple.
    """

    def __init__(self, *args, default=True, **kwargs):
        # Set a default to False
        self.default = default
        self.root = SPRUnOp(self.default)
        self.__overload_args_kwargs(*args, **kwargs)

    def __overload_args_kwargs(self, *args, **kwargs):
        if "scope" in kwargs:
            scope = kwargs["scope"]
            verb = kwargs.get("verb", None)
            self.root = SPRUnOp(ScopedPermissionRequirement(scope, verb))
        elif len(args) == 1:
            if "verb" in kwargs and isinstance(args[0], str):
                self.root = SPRUnOp(ScopedPermissionRequirement(args[0], args[1]))
            else:
                self.root = self._get_overloaded_arg(args[0])
        elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[1], str):
            self.root = SPRUnOp(ScopedPermissionRequirement(args[0], args[1]))
        else:
            self.root = self._get_overloaded_arg(args[0])

            for arg in args[1:]:
                self.root = SPROr(self.root, self._get_overloaded_arg(arg))

    def _get_overloaded_arg(self, arg):
        if isinstance(arg, str):
            return SPRUnOp(ScopedPermissionRequirement(arg))
        elif isinstance(arg, (tuple, list)) and len(arg) == 2:
            return SPRUnOp(ScopedPermissionRequirement(arg[0], arg[1]))
        elif isinstance(arg, ScopedPermissionRequirement):
            return SPRUnOp(arg)
        elif isinstance(arg, ScopedPermissionGuard):
            return arg.root
        elif isinstance(arg, (SPRBinOp, SPRUnOp)):
            return arg
        elif isinstance(arg, (tuple, list)):
            if len(arg) == 0:
                return SPRUnOp(self.default)

            base = self._get_overloaded_arg(arg[0])

            for arg in arg[1:]:
                base = SPROr(base, self._get_overloaded_arg(arg))

            return base
        else:
            return SPRUnOp(self.default)

    def has_permission(self, granting_scopes: Union[List[str], str], context=None):
        if isinstance(granting_scopes, str):
            granting_scopes = [granting_scopes]
        return self.root.has_permission(granting_scopes, context)

    def __and__(self, other):
        # Always create a new instance
        guard = ScopedPermissionGuard(self.root)
        guard.root = SPRAnd(guard.root, other.root)
        return guard

    def __or__(self, other):
        # Always create a new instance
        guard = ScopedPermissionGuard(self.root)
        guard.root = SPROr(guard.root, other.root)
        return guard

    def __xor__(self, other):
        # Always create a new instance
        guard = ScopedPermissionGuard(self.root)
        guard.root = SPRXor(guard.root, other.root)
        return guard

    def __invert__(self):
        # Always create a new instance
        guard = ScopedPermissionGuard(self.root)
        guard.root = SPRNot(guard.root)
        return guard

    def __len__(self):
        # Temporary to work around graphene-django-cuds length requirement for permissions
        return 1



from functools import wraps

from django.core.exceptions import PermissionDenied

from django_scoped_permissions.guards import ScopedPermissionGuard


def gql_has_scoped_permissions(
    *args,
    fail_message: str = "You are not permitted to view this",
    **kwargs,
):
    """
    gql_has_permissions is a function which wraps a `resolve_<x>` or
    `mutate` field for any GraphQL object.

    When called, it checks whether or not the calling user has permission to
    authorize the resource being requested, depending on the permissions given
    as necessary to access the resource.

    :param permissions: The permission required to access the wrapped resource.
    :param fail_message: If fail_to_none is false, and the permission fails, this variable determines
                         the string which is thrown in the exception.
    :return:
    """

    guard = ScopedPermissionGuard(*args, **kwargs)

    def decorator(func):
        @wraps(func)
        def wrapper(cls, info, *args, **kwargs):
            if not hasattr(info, "context") or not hasattr(info.context, "user"):
                raise PermissionDenied(fail_message)

            user = info.context.user
            if not user or user.is_anonymous:
                raise PermissionDenied(fail_message)

            context = {}
            context["context"] = info.context
            context["user"] = info.context.user

            if not guard.has_permission(user.get_granting_scopes(), context):
                raise PermissionDenied(fail_message)

            return func(cls, info, *args, **kwargs)

        return wrapper

    return decorator


def function_has_scoped_permissions(
    *args,
    fail_message: str = "You are not permitted to view this",
    **kwargs,
):
    guard = ScopedPermissionGuard(*args, **kwargs)

    def decorator(func):
        @wraps(func)
        def wrapper(request, *args, **kwargs):
            if not hasattr(request, "user"):
                raise PermissionDenied(fail_message)

            user = request.user
            if not user or user.is_anonymous:
                raise PermissionDenied(fail_message)

            context = {}
            context["context"] = request
            context["user"] = request.user

            if not guard.has_permission(user.get_granting_scopes(), context):
                raise PermissionDenied(fail_message)

            return func(request, *args, **kwargs)

        return wrapper

    return decorator



from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import AbstractUser

from django_scoped_permissions.models import ScopedPermissionHolder, ScopedModelMixin


class ScopedAuthenticationBackend(ModelBackend):
    def has_perm(self, user_obj: AbstractUser, perm: str, obj=None):
        if user_obj.is_anonymous:
            return False

        if user_obj.is_superuser:
            return True

        if not isinstance(user_obj, ScopedPermissionHolder):
            return None

        if not obj:
            return user_obj.has_scoped_permissions(perm)
        elif isinstance(obj, ScopedModelMixin):
            return obj.has_permission(user_obj, perm)
        else:
            return None



from django.test import TestCase

from django_scoped_permissions.models import ScopedPermission, ScopedPermissionGroup
from django_scoped_permissions.tests.factories import UserFactory
from django_scoped_permissions.tests.models import User


class TestHasScopedPermissionMixin(TestCase):
    def test_resolve_scopes__from_direct_m2m_field__resolves_correctly(self):
        user = UserFactory.create()

        scoped_permission_one = ScopedPermission.objects.create(
            scope="scope1"
        )
        scoped_permission_two = ScopedPermission.objects.create(
            scope="scope1:scope2"
        )
        scoped_permission_three = ScopedPermission.objects.create(
            scope="scope3:scope4",
            exact=True
        )
        scoped_permission_four = ScopedPermission.objects.create(
            scope="scope5:scope6",
            exclude=True
        )
        scoped_permission_five = ScopedPermission.objects.create(
            scope="scope7:scope8",
            exclude=True,
            exact=True
        )
        user.scoped_permissions.add(scoped_permission_one)
        user.scoped_permissions.add(scoped_permission_two)
        user.scoped_permissions.add(scoped_permission_three)
        user.scoped_permissions.add(scoped_permission_four)
        user.scoped_permissions.add(scoped_permission_five)

        resolved = user.resolved_scopes
        self.assertIn("scope1", resolved)
        self.assertIn("scope1:scope2", resolved)
        self.assertIn("=scope3:scope4", resolved)
        self.assertIn("-scope5:scope6", resolved)
        self.assertIn("-=scope7:scope8", resolved)

    def test_resolve_scopes__from_m2m_and_group_field__resolves_correctly(self):
        user = UserFactory.create()

        scoped_permission_one = ScopedPermission.objects.create(
            scope="scope1"
        )
        scoped_permission_two = ScopedPermission.objects.create(
            scope="scope1:scope2"
        )
        scoped_permission_three = ScopedPermission.objects.create(
            scope="scope3:scope4",
            exact=True
        )
        scoped_permission_four = ScopedPermission.objects.create(
            scope="scope5:scope6",
            exclude=True
        )
        scoped_permission_five = ScopedPermission.objects.create(
            scope="scope7:scope8",
            exclude=True,
            exact=True
        )
        user.scoped_permissions.add(scoped_permission_one)
        user.scoped_permissions.add(scoped_permission_two)

        group = ScopedPermissionGroup.objects.create(name="group")
        group.scoped_permissions.add(scoped_permission_three)
        group.scoped_permissions.add(scoped_permission_four)
        group.scoped_permissions.add(scoped_permission_five)
        user.scoped_permission_groups.add(group)

        resolved = user.resolved_scopes
        self.assertIn("scope1", resolved)
        self.assertIn("scope1:scope2", resolved)
        self.assertIn("=scope3:scope4", resolved)
        self.assertIn("-scope5:scope6", resolved)
        self.assertIn("-=scope7:scope8", resolved)

    def test_add_or_create_permission__permission_exists__adds_permissions(self):
        user: User = UserFactory.create()
        user.add_or_create_permission("scope1:scope2")
        self.assertTrue(user.has_any_scoped_permissions("scope1:scope2"))

        user.add_or_create_permission("=scope3:scope4")
        print(user.resolved_scopes)
        self.assertTrue(user.has_any_scoped_permissions("scope3:scope4"))
        self.assertFalse(user.has_any_scoped_permissions("scope3:scope4:test"))

        user.add_or_create_permission("scope3")
        user.add_or_create_permission("-scope3:scope4")
        self.assertTrue(user.has_any_scoped_permissions("scope3:scope5"))
        self.assertFalse(user.has_any_scoped_permissions("scope3:scope4:scope5"))

    def test_add_or_create_permission__permission_exists__reuses(self):
        permission_1 = ScopedPermission.objects.create(
            scope="scope1:scope2"
        )
        self.assertEqual(1, ScopedPermission.objects.count())
        user: User = UserFactory.create()
        user.add_or_create_permission("scope1:scope2")
        self.assertEqual(1, ScopedPermission.objects.count())

    def test_add_or_create_permission__permission_does_not_exist__creates_new(self):
        self.assertEqual(0, ScopedPermission.objects.count())
        user: User = UserFactory.create()
        user.add_or_create_permission("scope1:scope2")
        self.assertEqual(1, ScopedPermission.objects.count())
        permission = ScopedPermission.objects.filter(scope="scope1:scope2").first()
        self.assertIsNotNone(permission)




import graphene
from graphene import Node
from graphene_django import DjangoObjectType, DjangoConnectionField
from graphql import GraphQLError

from django_scoped_permissions.core import create_scope
from django_scoped_permissions.graphql import (
    ScopedDjangoCreateMutation,
    ScopedDjangoPatchMutation,
    ScopedDjangoUpdateMutation,
    ScopedDjangoDeleteMutation,
)
from django_scoped_permissions.tests.models import Pet, User


class PetNode(DjangoObjectType):
    class Meta:
        model = Pet
        interfaces = (Node,)

    @classmethod
    def get_node(self, info, id):
        return Pet.objects.get(pk=id)


class UserNode(DjangoObjectType):
    class Meta:
        model = User
        interfaces = (Node,)


class PetQuery(graphene.ObjectType):

    user = Node.Field(UserNode)

    pet = Node.Field(PetNode)
    all_pets = DjangoConnectionField(PetNode)

    def resolve_all_pets(self, info, *args, **kwargs):
        user = info.context.user
        company = info.context.company
        return Pet.objects.all()


class CreatePetMutation(ScopedDjangoCreateMutation):
    class Meta:
        model = Pet

        permissions = [
            create_scope(Pet, "create"),
            create_scope(User, "edit"),
        ]


class PatchPetMutation(ScopedDjangoPatchMutation):
    class Meta:
        model = Pet


class UpdatePetMutation(ScopedDjangoUpdateMutation):
    class Meta:
        model = Pet


class DeletePetMutation(ScopedDjangoDeleteMutation):
    class Meta:
        model = Pet


class CreateUserMutation(ScopedDjangoCreateMutation):
    class Meta:
        model = User


class PetMutations(graphene.ObjectType):
    create_pet = CreatePetMutation.Field()
    update_pet = UpdatePetMutation.Field()
    patch_pet = PatchPetMutation.Field()
    delete_pet = DeletePetMutation.Field()

    create_user = CreateUserMutation.Field()


class Query(PetQuery):
    pass


class CreateUserMutation(ScopedDjangoCreateMutation):
    class Meta:
        model = User
        optional_fields = ("password", "username")
        required_fields = ("primary_organization",)
        permissions = [
            create_scope(User, "create"),
        ]

    @classmethod
    def before_save(cls, root, info, input, obj):
        raise GraphQLError("Epic fail")
        obj.username = obj.email


class Mutation(graphene.ObjectType):
    create_user = CreateUserMutation.Field()


schema = graphene.Schema(query=Query, mutation=Mutation,)



# Generated by Django 3.0.3 on 2020-02-21 04:51

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ScopedPermission',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('scope', models.TextField()),
                ('exclude', models.BooleanField(default=False, help_text="Whether this should be an exclusive permission, meaning that if scope is 'user:update' and exclude is True then users with this usertype will not be able to update users, even their own.")),
                ('exact', models.BooleanField(default=False, help_text='If checked, the permission needs an exact match to count. In other words, it does not work recursively as standard scoped permissions.')),
            ],
            options={
                'unique_together': {('scope', 'exclude', 'exact')},
            },
        ),
    ]
